/*****************************************************************************

	unsort - reorder files semi-randomly
	Copyright (C) 2007  Wessel Dankers <wsl@fruit.je>

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

*****************************************************************************/

#ifndef MT19937AR_H
#define MT19937AR_H

#include <stdint.h>

extern void mt_seed(uint32_t);
extern void mt_seed_array(uint32_t *, unsigned int);
extern int32_t mt_genrand31(void);
extern uint32_t mt_genrand32(void);
extern double mt_genrand_real1(void);
extern double mt_genrand_real2(void);
extern double mt_genrand_real3(void);
extern double mt_genrand_res53(void);
extern uint32_t mt_genrand32_bounded(uint32_t min, uint32_t max);

#endif
